function [fhat,time]=multi_lattice_ifft_noniter_avg(f,h,Ms,zs)
%MULTI_LATTICE_IFFT_NONITER_AVG performs FFT for multiple reconstructing rank-1 lattice configuration,
%e.g., constructed by multi_lattice_construct_mr1l_2.m
%
%   See also Algorithm 2 in
%      L. Kaemmerer, D. Potts, and T. Volkmer.
%      High-dimensional sparse FFT based on sampling along multiple rank-1 lattices.
%      ArXiv e-prints 1711.05152, 2017.
%
%   Input parameters:
%      f - sampling values
%      h - frequency set
%      Ms - lattice sizes
%      zs - generating vectors
%
%   Output parameters:
%      fhat - Fourier coefficients
%
% Copyright (c) 2017-2020 Lutz Kaemmerer, Toni Volkmer
%
% This program is free software; you can redistribute it and/or modify it under
% the terms of the GNU General Public License as published by the Free Software
% Foundation; either version 2 of the License, or (at your option) any later
% version.
%
% This program is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
% FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
% details.
%
% You should have received a copy of the GNU General Public License along with
% this program; if not, write to the Free Software Foundation, Inc., 51
% Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

    ttimer = tic;

    fhat = zeros(size(h,1),1);
    counter = zeros(size(h,1),1);

    for j=1:length(Ms)
      y = mod(h*zs(j,:)',Ms(j)) + 1;
      num_1d = accumarray(y,1,[Ms(j),1],@sum);
      num_alias = num_1d(y);
      ind_reco = find(num_alias == 1);
      
      if j==1
        f1=f(1:Ms(1));
      else
        f1=[f(1);f(sum(Ms(1:j-1))-j+3:sum(Ms(1:j))-j+1)];
      end

      a1 = 1/Ms(j)*lfft(h(ind_reco,:),zs(j,:),Ms(j),f1,'transp');
      fhat(ind_reco) = fhat(ind_reco) + a1;
      counter(ind_reco) = counter(ind_reco) + 1;
    end

    fhat = fhat ./ counter;

    time = toc(ttimer);
end
