function [ num_freq_out, err_l2_rel, err_l2_rel_trunc, err_l2_rel_alias, freq_cand_sizes, lattice_sizes, sample_sizes, z, run_times, niter_cg ] = test3_1_rr1l_sfft_algorithm_singletest( N, algorithm_name, sparsity_s, sparsity_s_local, threshold_theta, threshold_theta_basic, niter_r, niter_r_basic )
%TEST3_1_RR1L_SFFT_ALGORITHM_SINGLETEST Perform single test from Section 4.2.2.
%   A (fixed) 10-dimensional test function with infinitely many non-zero
%   Fourier coefficents is used, which consists of tensor products
%   of several B-splines of different order.
%   For given relative threshold values, approximately the largest
%   Fourier coefficients with corresponding frequencies are reconstructed
%   by Algorithm 3 using Algorithm 4.
%
%   Input parameters:
%      N - expansion parameter
%      algorithm_name - use 'rlfftriabs'
%      sparsity_s - sparsity parameter
%      sparsity_s_local - internal parameter
%      threshold_theta - internal parameter
%      threshold_local - internal parameter
%      niter_r - number of detection iterations
%      niter_r_basic - number of detection iterations
%
%   Output Parameters:
%      num_freq_out - number of detected frequencies
%      err_l2_rel - relative L2 (sampling) error
%      err_l2_rel_trunc - relative L2 truncation/approximation error
%      err_l2_rel_alias - relative L2 aliasing error
%      freq_cand_sizes - internal value
%      lattice_sizes - number of lattices internally used
%      sample_sizes - number of samples
%      z - empty array
%      run_times - algorithm runtimes
%      niter_cg - empty array
%
% Copyright (c) 2020 Toni Volkmer
%
% This program is free software; you can redistribute it and/or modify it under
% the terms of the GNU General Public License as published by the Free Software
% Foundation; either version 2 of the License, or (at your option) any later
% version.
%
% This program is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
% FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
% details.
%
% You should have received a copy of the GNU General Public License along with
% this program; if not, write to the Free Software Foundation, Inc., 51
% Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

if nargin < 8
  niter_r_basic = niter_r;
end

switch algorithm_name
  case {'rlfftriabs'}
    prob_delta = -0.25; % negative value -0.25 means 1/4 of theoretical number
    [freq_out, fhat_out, freq_cand_sizes, num_R1L_sizes, sample_sizes, run_times] ...
      = random_lattice_sfft_algorithm(@bspline_test_10d, 10, N, algorithm_name, sparsity_s, sparsity_s_local, threshold_theta, threshold_theta_basic, niter_r, niter_r_basic, prob_delta );
    lattice_sizes = num_R1L_sizes;
    z = [];
    niter_cg = [];
  otherwise
    error('Algorithm not supported');
end

fhat = bspline_test_10d_fouriercoeff(freq_out);

num_freq_out = size(freq_out,1);

%norm_fct_square = 3 + 2*bspline_o2_hat(zeros(1,3))*bspline_o4_hat(zeros(1,4)) + 2*bspline_o2_hat(zeros(1,3))*bspline_o6_hat(zeros(1,3)) + 2*bspline_o4_hat(zeros(1,4))*bspline_o6_hat(zeros(1,3));
norm_fct_square = 3.86052137015856367;
fprintf('norm fct: %.6e, ', sqrt(norm_fct_square));
err_l2_rel_trunc = sqrt(norm_fct_square - norm(fhat,2)^2) / sqrt(norm_fct_square);
fprintf('rel. truncation err.: %.3e, ', err_l2_rel_trunc);
err_l2_rel_alias = norm(fhat_out-fhat,2) / sqrt(norm_fct_square);
fprintf('rel. aliasing err.: %.3e\n', err_l2_rel_alias);

err_l2_rel = sqrt( norm_fct_square - norm(fhat,2)^2 + norm(fhat_out-fhat,2)^2 ) / sqrt(norm_fct_square);
fprintf('relative L_2 approximation error: %3e\n', err_l2_rel);

end
